unit GLRender;

// 3D object OpenGL Render (c)2001 by Paul TOTH <tothpaul@free.fr>

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
interface
}

{$define cull}
interface

uses
  Classes,SysUtils,DelphiGL,GLTextures;

type
  TVector=record
   x,y,z:single;
  end;
  TVertex=TVector;
  TVertices=array { VertexCount } of TVertex;

  TColor4f=record
   r,g,b,a:single;
  end;
  TColors=array of TColor4f;

  TFace=record
   a,b,c:integer;
   group:integer;
   material:integer;
  end;
  TFaces=array { FaceCount } of TFace;

  TTexture=class
   FileName:string;
   Handle  :integer;
   dx,dy   :single;
  end;
  TTextures=array of TTexture;

  TMaterial=record
   Name    :string;
   Ambient :integer;
   Diffuse :integer;
   Specular:integer;
   TwoSide :boolean;
   Transparency:single;
   Texture :integer;
  end;
  TMaterials=array of TMaterial;

  TMapping=array[0..2] of record
   u,v:single;
  end;
  TMappings=array { FaceCount } of TMapping;

  TFrame=record
   Name    :string;
   Vertices:TVertices;
   Faces   :TFaces;
  end;
  TFrames=array of TFrame;

  TGLObject = class
  private
   fVertices :TVertices;
   fFaces    :TFaces;
   fFullDetail:TFaces;
   fLOD      :integer;
   fPoints   :TVertices;
   fTextures :TTextures;
   fMappings :TMappings;
   fFNormals :array { FaceCount  } of record x,y,z,fz:single end;
   fGouraud  :array { VertexCount} of record r,g,b,a:single end;
   fVNormals :array { VertexCount} of record x,y,z:single end;
   fZSort    :array { FaceCount  } of integer;
   fColors   :TColors;
   fMaterials:TMaterials;
   fGroups   :TStringList;
   fFaceCount:integer;
   fVertexCount:integer;
   fTransparent:boolean;
   fFrameCount:integer;
   fFrames:TFrames;
   fFrame:integer;
   function GetMaterialCount:integer;
   function GetMaterial(Index:integer):TMaterial;
   procedure SetMaterial(Index:integer; Value:TMaterial);
   function GetColor(Index:integer):TColor4f;
   procedure SetColor(Index:integer; Value:TColor4f);
   function GetTexture(Index:integer):TTexture;
   procedure SetTexture(Index:integer; Value:TTexture);
   procedure SetFrame(Value:integer);
   function GetFrameName(index:integer):string;
  protected
   procedure ComputeNormals;
   procedure ZSort;
  public
   Constructor Create;
   Destructor Destroy; override;
   procedure Clear;
   procedure LoadFromFile(FileName:string);
   procedure Normalise;
   procedure Reduice(AVertexCount:integer);
   procedure Reset;
   procedure Rotate(rx,ry,rz:single);
   procedure Translate(tx,ty,tz:single);
   procedure Render;
   procedure GLRender;
   procedure RenderSelection(Index:integer);
   function AddTexture(AFileName:string):integer;
   procedure LoadTextures;
   procedure FreeTextures;
   property Groups:TStringList read fGroups;
   property VertexCount:integer read fVertexCount;
   property FaceCount:integer read fFaceCount;
   property MaterialCount:integer read GetMaterialCount;
   property Materials[Index:integer]:TMaterial read GetMaterial write SetMaterial;
   property Colors[Index:integer]:TColor4f read GetColor write SetColor;
   property Textures[Index:integer]:TTexture read GetTexture write SetTexture;
   property FrameCount:integer read fFrameCount;
   property Frame:integer read fFrame write SetFrame;
   property FrameName[index:integer]:string read GetFrameName;
  end;

  TGLReader=class
   Vertices   :TVertices;
   Faces      :TFaces;
   Groups     :TStrings;
   Colors     :TColors;
   Transparent:boolean;
   Materials  :TMaterials;
   Mappings   :TMappings;
   Textures   :TTextures;
   FrameCount :integer;
   procedure NextFrame; virtual;
  end;

{$ifdef log}
type
 TLog=procedure(s:string) of object;
var
 Log:TLog;
{$endif}

implementation

Uses
 Load3DS, // 3D Studio           .3DS
 LoadOBJ, // Poser wavefront     .OBJ
 LoadCR2, // Poser Figure        .CR2
 LoadMD2, // Quake 2 Model       .MD2
 LoadCFG, // Cal3D configuration .CFG
 ProgMesh;

procedure TGLReader.NextFrame;
 begin
  raise Exception.Create('Can''t load next frame');
 end;

Constructor TGLObject.Create;
 begin
  fGroups:=TStringList.Create;
 end;

Destructor TGLObject.Destroy;
 begin
  Clear;
  fGroups.Free;
  inherited;
 end;

Procedure TGLObject.Clear;
 var
  i:integer;
 begin
  for i:=0 to Length(fTextures)-1 do begin
   with fTextures[i] do begin
    if Handle<>-1 then glDeleteTextures(1,Handle);
    Free;
   end;
  end;
  fVertices:=nil;
  fFaces:=nil;
  fColors:=nil;
  fMaterials:=nil;
  fMappings:=nil;
  fTextures:=nil;
  fGroups.Clear;
  fVertexCount:=0;
  fFaceCount:=0;
  fTransparent:=False;
  fZSort:=nil;
  fPoints:=nil;
 end;

//----------------------------------------------------------------------------//
Procedure TGLObject.Normalise;
 var
  xmin,xmax:single;
  ymin,ymax:single;
  zmin,zmax:single;
  oldf,f,i:integer;
  z1,z2:single;
  ox,oy,oz:single;
 begin
  if (fVertexCount=0)or(fFaceCount=0) then exit;
  xmin:=fVertices[0].x; xmax:=xmin;
  ymin:=fVertices[0].y; ymax:=ymin;
  zmin:=fVertices[0].z; zmax:=zmin;
  oldf:=Frame;
  for f:=0 to fFrameCount-1 do begin
   Frame:=f;
   for i:=1 to fVertexCount-1 do with fVertices[i] do begin
    if x<xmin then xmin:=x else if x>xmax then xmax:=x;
    if y<ymin then ymin:=y else if y>ymax then ymax:=y;
    if z<zmin then zmin:=z else if z>zmax then zmax:=z;
   end;
  end;
  z1:=xmax-xmin;
  z2:=ymax-ymin; if z2>z1 then z1:=z2;
  z2:=zmax-zmin; if z2>z1 then z1:=z2;
  ox:=xmin+(xmax-xmin)/2;
  oy:=ymin+(ymax-ymin)/2;
  oz:=zmin+(zmax-zmin)/2;
  for f:=0 to fFrameCount-1 do begin
   Frame:=f;
   for i:=0 to fVertexCount-1 do with fVertices[i] do begin
    x:=(x-ox)/z1;
    y:=(y-oy)/z1;
    z:=(z-oz)/z1;
   end;
  end;
  Frame:=Oldf;
  Reset;
 end;

procedure TGLObject.Reduice(AVertexCount:integer);
 begin
  fLOD:=AVertexCount;
  fFaces:=ProgressiveMesh(fVertices,fFullDetail,AVertexCount);
  fFaceCount:=Length(fFaces);
 end;

procedure TGLObject.Reset;
 begin
  Move(fVertices[0],fPoints[0],fVertexCount*SizeOf(TVertex));
 end;

Procedure TGLObject.Translate(tx,ty,tz:single);
 var
  i:integer;
 begin
  for i:=0 to fVertexCount-1 do begin
   with fPoints[i] do begin
    x:=x+tx;
    y:=y+ty;
    z:=z+tz;
   end;
  end;
 end;

Procedure TGLObject.Rotate(rx,ry,rz:single);
 var
  cosx,cosy,cosz,sinx,siny,sinz:single;
  rx1,rx2,rx3:single;
  ry1,ry2,ry3:single;
  rz1,rz2,rz3:single;
  v:TVertex;
  i:integer;
 begin
  rx:=rx*PI/180;
  ry:=ry*PI/180;
  rz:=rz*PI/180;
  cosx:=cos(rx);
  cosy:=cos(ry);
  cosz:=cos(rz);
  sinx:=sin(rx);
  siny:=sin(ry);
  sinz:=sin(rz);
 // rotation matrix
  rx1:= cosy*cosz; rx2:=sinx*siny*cosz-cosx*sinz; rx3:=cosx*siny*cosz+sinx*sinz;
  ry1:= cosy*sinz; ry2:=sinx*siny*sinz+cosx*cosz; ry3:=cosx*siny*sinz-sinx*cosz;
  rz1:=-siny;      rz2:=sinx*cosy;                rz3:=cosx*cosy;
  SetLength(fPoints,fVertexCount);
  for i:=0 to fVertexCount-1 do begin
   v:=fPoints[i];
   with fPoints[i] do begin
    x:=rx1*v.x+rx2*v.y+rx3*v.z;
    y:=ry1*v.x+ry2*v.y+ry3*v.z;
    z:=rz1*v.x+rz2*v.y+rz3*v.z;
   end;
  end;
 end;

Procedure TGLObject.ComputeNormals;
 type
  TColorInfo=record
   count:integer;
   r,g,b:single;
   a:single;
   ncount:integer;
   x,y,z:single;
  end;
 var
  n:integer;
  x1,y1,z1,x2,y2,z2,vl,nx,ny,nz:single;
  ColorInfo:array of Tcolorinfo;

  procedure SetColorInfo(index,Material:integer);
   var
    i:integer;
   begin
    with ColorInfo[Index] do begin
     inc(ncount);
     x:=x+nx;
     y:=y+ny;
     z:=z+nz;
     if (Material<0) then begin
      inc(count);
      r:=r+nz;
      g:=g+nz;
      b:=b+nz;
     end else begin
      a:=a+nz*fMaterials[Material].Transparency;
      i:=fMaterials[Material].Diffuse;
      if (i>=0) then begin
       //inc(count);
       r:=r+nz*fColors[i].r;
       g:=g+nz*fColors[i].g;
       b:=b+nz*fColors[i].b;
      end;
      i:=fMaterials[Material].Ambient;
      if i>=0 then begin
       inc(count);
       r:=r+(1-nz)*fColors[i].r;
       g:=g+(1-nz)*fColors[i].g;
       b:=b+(1-nz)*fColors[i].b;
      end;
      i:=fMaterials[Material].Specular;
      if i>=0 then begin
       inc(count);
       r:=r+nz*fColors[i].r;
       g:=g+nz*fColors[i].g;
       b:=b+nz*fColors[i].b;
      end;
     end;
    end;
   end;

 begin
 // faces normals
  SetLength(fFNormals,fFaceCount);
  SetLength(fVNormals,fVertexCount);
  SetLength(ColorInfo,fVertexCount);
  for n:=0 to fFaceCount-1 do with fFaces[n] do begin
   fFNormals[n].fz:=fPoints[a].z+fPoints[b].z+fPoints[c].z;
   x1:=fPoints[b].x-fPoints[a].x;
   y1:=fPoints[b].y-fPoints[a].y;
   z1:=fPoints[b].z-fPoints[a].z;
   x2:=fPoints[c].x-fPoints[a].x;
   y2:=fPoints[c].y-fPoints[a].y;
   z2:=fPoints[c].z-fPoints[a].z;
   nx:=y1*z2-y2*z1;
   ny:=z1*x2-z2*x1;
   nz:=x1*y2-x2*y1;
   vl:=sqrt(nx*nx+ny*ny+nz*nz);
   if vl<>0 then begin
    nx:=nx/vl;
    ny:=ny/vl;
    nz:=nz/vl;
   end;
   if (Material<0)or(fMaterials[Material].TwoSide) then begin
    nx:=abs(nx);
    ny:=abs(ny);
    nz:=abs(nz);
   end;
   fFNormals[n].x:=nx;
   fFNormals[n].y:=ny;
   fFNormals[n].z:=nz;
   {$ifdef cull}if nz>=0 then{$endif} begin
    SetColorInfo(a,Material);
    SetColorInfo(b,Material);
    SetColorInfo(c,Material);
   end;
  end;
 // vertices colors
  SetLength(fGouraud,fVertexCount);
  for n:=0 to fVertexCount-1 do begin
   with ColorInfo[n] do begin
    if ncount>0 then begin
     fVNormals[n].x:=x/ncount;
     fVNormals[n].y:=y/ncount;
     fVNormals[n].z:=z/ncount;
    end;
    if count>0 then begin
     fGouraud[n].r:=r/count;
     fGouraud[n].g:=g/count;
     fGouraud[n].b:=b/count;
     fGouraud[n].a:=a/count;
    end;
   end;
  end;
 end;

procedure TGLObject.ZSort;

 procedure QuickSort(il,ir:integer);
  var
   l,m,r,t:integer;
   z:single;
  begin
   l:=il;
   r:=ir;
   m:=(l+r) div 2;
   z:=fFNormals[fZSort[m]].fz;
   repeat
    while fFNormals[fZSort[l]].fz<z do inc(l);
    while fFNormals[fZSort[r]].fz>z do dec(r);
    if l<=r then begin
     t:=fZSort[l];
     fZSort[l]:=fZSort[r];
     fZSort[r]:=t;
     inc(l);
     dec(r);
    end;
   until l>r;
   if r>il then QuickSort(il,r);
   if l<ir then Quicksort(l,ir);
  end;

 var
  i:integer;
 begin
  SetLength(fZSort,fFaceCount);
  for i:=0 to fFaceCount-1 do fZSort[i]:=i;
  QuickSort(0,fFaceCount-1);
 end;

Procedure TGLObject.Render;
 var
  i,s,t:integer;
  w,h:single;
 begin
  if fFaceCount=0 then exit;

  ComputeNormals;

  if fTransparent then begin
   ZSort;
   glEnable(GL_BLEND);
   glBlendFunc(GL_ONE_MINUS_SRC_ALPHA,GL_SRC_ALPHA);
  end else begin
   glDisable(GL_BLEND);
  end;

  for i:=0 to fFaceCount-1 do begin

   s:=fZSort[i];
   {$ifdef cull}if fFNormals[s].z>=0 then{$endif}

    with fFaces[s] do begin

    if Material<0 then t:=-1 else t:=fMaterials[Material].Texture;

     glDisable(GL_LIGHTING);
     glDisable(GL_LIGHT0);

     if (t<0)or(fTextures[t].Handle<0) then begin
      glDisable(GL_TEXTURE_2D);
      glBegin(GL_TRIANGLES);
      with fGouraud[a] do glColor4f (r,g,b,a);
      with fPoints [a] do glVertex3f(x,y,z);
      with fGouraud[b] do glColor4f (r,g,b,a);
      with fPoints [b] do glVertex3f(x,y,z);
      with fGouraud[c] do glColor4f (r,g,b,a);
      with fPoints [c] do glVertex3f(x,y,z);
      glEnd;
     end else begin
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D,fTextures[t].Handle);
      w:=fTextures[t].dx;
      h:=fTextures[t].dy;
      with fFNormals[s] do glNormal3f(x,y,z);
      glBegin(GL_TRIANGLES);
      with fMappings[s,0] do glTexCoord2f(u*w,v*h);
      with fGouraud[a] do glColor4f (r,g,b,a);
      with fPoints [a] do glVertex3f(x,y,z);
      with fMappings[s,1] do glTexCoord2f(u*w,v*h);
      with fGouraud[b] do glColor4f (r,g,b,a);
      with fPoints [b] do glVertex3f(x,y,z);
      with fMappings[s,2] do glTexCoord2f(u*w,v*h);
      with fGouraud[c] do glColor4f (r,g,b,a);
      with fPoints [c] do glVertex3f(x,y,z);
      glEnd;
     end;

    end;

  end;

 end;

const
 shininess:single=50;

Procedure TGLObject.GLRender;
 var
  i,s,t:integer;
  w,h:single;
 begin
  if fFaceCount=0 then exit;

  ComputeNormals;

  if fTransparent then begin
   ZSort;
   glEnable(GL_BLEND);
   glBlendFunc(GL_ONE_MINUS_SRC_ALPHA,GL_SRC_ALPHA);
  end else begin
   glDisable(GL_BLEND);
  end;

  for i:=0 to fFaceCount-1 do begin

   s:=fZSort[i];
   {$ifdef cull}if fFNormals[s].z>=0 then{$endif}

    with fFaces[s] do begin

    if Material<0 then t:=-1 else t:=fMaterials[Material].Texture;

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    if (Material>=0) then begin

     with fMaterials[Material] do begin
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,  fColors[Ambient]);
      glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,  fColors[Diffuse]);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, fColors[Specular]);
     end;
     glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

    end;

    glBegin(GL_TRIANGLES);
     with fVNormals[a] do glNormal3f(x,y,z);
     with fPoints  [a] do glVertex3f(x,y,z);
     with fVNormals[b] do glNormal3f(x,y,z);
     with fPoints  [b] do glVertex3f(x,y,z);
     with fVNormals[c] do glNormal3f(x,y,z);
     with fPoints  [c] do glVertex3f(x,y,z);
    glEnd;


    end;

  end;

 end;

Procedure TGLObject.RenderSelection(Index:integer);
 var
  i:integer;
 begin
  if Index<0 then exit;
  glDisable(GL_DEPTH_TEST);
  glEnable(GL_BLEND);
  glBlendFunc(GL_ONE_MINUS_SRC_ALPHA,GL_SRC_ALPHA);
  glColor4f(0.8,0,0,0.75);
  glBegin(GL_TRIANGLES);
  for i:=0 to fFaceCount-1 do with fFaces[i] do begin
   if Group=Index then begin
    with fPoints [a] do glVertex3f(x,y,z);
    with fPoints [b] do glVertex3f(x,y,z);
    with fPoints [c] do glVertex3f(x,y,z);
   end;
  end;
  glEnd;
  glEnable(GL_DEPTH_TEST);
 end;

function TGLObject.GetMaterialCount:integer;
 begin
  Result:=Length(fMaterials);
 end;

function TGLObject.GetMaterial(Index:integer):TMaterial;
 begin
  Result:=fMaterials[Index];
 end;

procedure TGLObject.SetMaterial(Index:integer;Value:TMaterial);
 begin
  fMaterials[Index]:=Value;
 end;

function TGLObject.GetColor(Index:integer):TColor4f;
 begin
  result:=fColors[Index];
 end;

procedure TGLObject.SetColor(Index:integer; Value:TColor4f);
 begin
  fColors[Index]:=Value;
 end;

function TGLObject.AddTexture(AFileName:string):integer;
 var
  i:integer;
 begin
  for i:=0 to Length(fTextures)-1 do begin
   if fTextures[i].FileName=AFileName then begin
    Result:=i;
    exit;
   end;
  end;
  Result:=Length(fTextures);
  SetLength(fTextures,Result+1);
  fTextures[Result]:=TTexture.Create;
  fTextures[Result].FileName:=AFileName;
 end;

procedure TGLObject.LoadTextures;
 var
  i:integer;
 begin
  for i:=0 to Length(fTextures)-1 do begin
   with fTextures[i] do begin
    try
     Handle:=LoadTexture(FileName,dx,dy);
    except
     Handle:=-1;
    end;
   end;
  end;
 end;

procedure TGLObject.FreeTextures;
 var
  i:integer;
 begin
  for i:=0 to Length(fTextures)-1 do begin
   with fTextures[i] do begin
    if Handle<>-1 then glDeleteTextures(1,Handle);
    Handle:=-1;
   end;
  end;
 end;

function TGLObject.GetTexture(Index:integer):TTexture;
 begin
  Result:=fTextures[index];
 end;

procedure TGLObject.SetTexture(Index:integer; Value:TTexture);
 begin
  fTextures[index]:=Value;
 end;

//----------------------------------------------------------------------------//
Procedure TGLObject.LoadFromFile(FileName:string);
 var
  ext:string;
  i:integer;
  Reader:TGLReader;
 begin
  ext:=Uppercase(ExtractFileExt(FileName));
  if ext='.3DS' then Reader:=T3DSReader.Create(FileName) else
  if ext='.OBJ' then Reader:=TOBJReader.Create(FileName) else
  if ext='.CR2' then Reader:=TCR2Reader.Create(FileName) else
  if ext='.MD2' then Reader:=TMD2Reader.Create(FileName) else raise Exception.Create('unknow file extension');
  Clear;
  fVertices:=reader.Vertices;
  fFaces:=reader.Faces;
  fFullDetail:=fFaces;
  fColors:=reader.Colors;
  fMaterials:=reader.Materials;
  fMappings:=reader.Mappings;
  fGroups.Assign(reader.Groups);
  fVertexCount:=Length(fVertices);
  fFaceCount:=Length(fFaces);
  fTransparent:=reader.Transparent;
  fTextures:=reader.Textures;

  fFrameCount:=reader.FrameCount;
  if fFrameCount>1 then begin
   SetLength(fFrames,fFrameCount);
   i:=0;
   repeat
    if i>0 then reader.NextFrame;
    fFrames[i].Name:=reader.Groups[0];
    fFrames[i].Vertices:=reader.Vertices;
    fFrames[i].Faces:=reader.Faces;
    inc(i);
   until i=fFrameCount;
  end;

  Reader.Free;
  SetLength(fZSort,fFaceCount);
  for i:=0 to fFaceCount-1 do fZSort[i]:=i;
  SetLength(fPoints,fVertexCount);

  SetFrame(0);
 end;

Procedure TGLObject.SetFrame(Value:integer);
 begin
  if Value>=fFrameCount then fFrame:=0 else fFrame:=Value;
  if fFrameCount<=1 then exit;
  fVertices:=fFrames[fFrame].Vertices;
  fFaces:=fFrames[fFrame].Faces;
  Reduice(fLOD);
 end;

Function TGLObject.GetFrameName(index:integer):string;
 begin
  result:=fFrames[Index].Name;
 end;

end.






