unit Load3DS;

{$define log
  3DStudio Loader (C)2001 by Paul TOTH <tothpaul@free.fr>
  http://tothpaul.free.fr
}

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
interface
}

interface

Uses
 Classes, SysUtils, GLRender;
//-- 3DS ---------------------------------------------------------------------//
Type
 T3DSChunk=packed record
  Tag:word;
  Size:integer
 end;

 T3DSColor=packed record
  r,g,b:single;
 end;

 T3DSFace=packed record
  a,b,c,flag:word;
 end;

 T3DSMapping=packed record
  u,v:single;
 end;
 T3DSMappings=array { VertexCount } of T3DSMapping;

 T3DSMeshMatrix=packed record
  unknow1 :TVector;
  unknow2 :TVector;
  unknow3 :TVector;
  Position:TVector;
 end;

 T3DSTrackHeader=packed record
  Flags  :word;
  unused1:integer;
  unused2:integer;
  Count  :integer;
 end;

 T3DSKeyHeader=packed record
  Time      :integer;
  Flags     :word;
  Tension   :single; // if (Flags and  1)<>0
  Continuity:single; // if (Flags and  2)<>0
  Bias      :single; // if (Flags and  4)<>0
  EaseTo    :single; // if (Flags and  8)<>0
  EaseFrom  :single; // if (Flags and 16)<>0
 end;

 T3DSReader=class(TGLReader)
 private
 {$ifdef log}
  Iter         :string;
 {$endif}
  Stream       :TStream;
  Chunk        :T3DSChunk;
  VertexCount  :integer;
  FaceCount    :integer;
  MaterialCount:integer;
  ColorCount   :integer;
  TextureCount :integer;
  Group        :integer;
  Vertex0      :integer;
  face0        :integer;
  fMappings    :T3DSMappings;
  Pivot        :TVertex;
  Matrix       :T3DSMeshMatrix;
  function LoadChunk(Var Size:integer):word;
  procedure Load(Name:string);
  function LoadWord:word;
  function LoadSingle:single;
  function LoadString:string;
  function LoadByteRGB:integer;
  function LoadFloatRGB:integer;
  procedure LoadMaterial;
  function LoadColor:integer;
  function LoadTransparency:single;
  Procedure LoadObject;
  procedure LoadVertices;
  procedure LoadFaces;
  procedure LoadMaterials;
  procedure LoadMappings;
  procedure LoadMeshMatrix;
  procedure LoadNodeHeader;
  procedure LoadKeyHeader(out Header:T3DSKeyHeader);
  procedure LoadPivot;
  procedure LoadTranslation;
  procedure LoadRotation;
  procedure LoadScale;
  procedure TranslateGroup(const v:TVector);
  function AddTexture(AFileName:string):integer;
 public
  Constructor Create(AFileName:string);
  Destructor Destroy; override;
 end;

implementation

Constructor T3DSReader.Create(AFileName:string);
 var
  i,m:integer;
 begin
  Stream:=TFileStream.Create(AFileName,fmOpenRead or fmShareDenyNone);
  Groups:=TStringList.Create;
  Mappings:=nil;
  i:=Stream.size;
  if LoadChunk(i)<>$4D4D then raise Exception.Create('Not a 3DS File');
  Load('3DS');
  m:=Length(fMappings);
  if m>0 then begin
   SetLength(Mappings,FaceCount);
   for i:=0 to FaceCount-1 do with Faces[i] do begin
    Mappings[i,0].u:=fMappings[a].u;
    Mappings[i,0].v:=fMappings[a].v;
    Mappings[i,1].u:=fMappings[b].u;
    Mappings[i,1].v:=fMappings[b].v;
    Mappings[i,2].u:=fMappings[c].u;
    Mappings[i,2].v:=fMappings[c].v;
   end;
  end;
  FrameCount:=1;
 end;

Destructor T3DSReader.Destroy;
 begin
  Stream.Free;
  Groups.Free;
 end;

function T3DSReader.LoadChunk(var Size:integer):word;
 begin
  Stream.ReadBuffer(Chunk,SizeOf(Chunk));
  Dec(Size,Chunk.Size);
  Result:=Chunk.Tag;
 end;

procedure T3DSReader.Load(Name:string);
 var
  size:integer;
 begin
  size:=Chunk.Size-sizeof(T3DSChunk);
  While Size>0 do begin
   case LoadChunk(Size) of
    $3D3D: Load('Mesh');
     $AFFF: LoadMaterial;
      $A000: Materials[MaterialCount].Name:=LoadString;
       $A010: Materials[MaterialCount].Ambient:=LoadColor;
       $A020: Materials[MaterialCount].Diffuse:=LoadColor;
       $A030: Materials[MaterialCount].Specular:=LoadColor;
       $A050: Materials[MaterialCount].Transparency:=LoadTransparency;
       $A081: Materials[MaterialCount].TwoSide:=True;
       $A200: Load('Texmap');
        $A300: Materials[MaterialCount].Texture:=AddTexture(LoadString);
     $4000: LoadObject;
      $4100: Load('Data');
       $4110: LoadVertices;  // vertices
       $4120: LoadFaces;     // faces
       $4130: LoadMaterials; // colors & textures..
       $4140: LoadMappings;
       $4160: LoadMeshMatrix;
(******)
     $B000: Load('KeyFrameData');
      $B002: Load('ObjectNode');
       $B010: LoadNodeHeader;
       $B013: LoadPivot;
       $B020: LoadTranslation;
       $B021: LoadRotation;
       $B022: LoadScale;
(******)
    else Stream.Position:=Stream.Position+Chunk.Size-SizeOf(T3DSChunk); // skip chunk
   end;
  end;
 end;

function T3DSReader.LoadWord:word;
 begin
  Stream.ReadBuffer(Result,sizeof(Result));
 end;

function T3DSReader.LoadSingle:single;
 begin
  Stream.ReadBuffer(Result,sizeof(Result));
  Dec(Chunk.Size,sizeof(Result));
 end;

function T3DSReader.LoadString:string;
 var
  i:integer;
 begin
  SetLength(result,255);
  i:=0;
  repeat
   inc(i);
   Stream.ReadBuffer(Result[i],1);
  until Result[i]=#0;
  SetLength(Result,i-1);
  Dec(Chunk.Size,i);
 end;

function T3DSReader.LoadByteRGB:integer;
 var
  rgb:packed record r,g,b:byte end;
  color:TColor4f;
  i:integer;
 begin
  Stream.ReadBuffer(rgb,sizeof(rgb));
  color.r:=rgb.r/256;
  color.g:=rgb.g/256;
  color.b:=rgb.b/256;
  color.a:=1;
  for i:=0 to ColorCount-1 do with Colors[i] do begin
   if (r=color.r)and(g=color.g)and(b=color.b) then begin
    result:=i;
    exit;
   end;
  end;
  SetLength(Colors,ColorCount+1);
  Colors[ColorCount]:=Color;
  result:=ColorCount;
  inc(ColorCount);
 end;

function T3DSReader.LoadFloatRGB:integer;
 var
  color:T3DSColor;
  i:integer;
 begin
  Stream.ReadBuffer(Color,sizeof(T3DSColor));
  for i:=0 to ColorCount-1 do with Colors[i] do begin
   if (r=color.r)and(g=color.g)and(b=color.b) then begin
    result:=i;
    exit;
   end;
  end;
  SetLength(Colors,ColorCount+1);
  Colors[ColorCount].r:=color.r;
  Colors[ColorCount].g:=color.g;
  Colors[ColorCount].b:=color.b;
  Colors[ColorCount].a:=1;
  result:=ColorCount;
  inc(ColorCount);
 end;

procedure T3DSReader.LoadMaterial;
 begin
  SetLength(Materials,MaterialCount+1);
  Materials[MaterialCount].Texture:=-1;
  Load('Material');
  if Materials[MaterialCount].Transparency>0 then Transparent:=True;
  Inc(MaterialCount);
 end;

function T3DSReader.LoadColor:integer;
 var
  size:integer;
 begin
  Result:=0;
  Size:=Chunk.Size-SizeOf(T3DSChunk);
  while Size>0 do begin
   case LoadChunk(Size) of
    $0010 : Result:=LoadFloatRGB;
    $0011 : Result:=LoadByteRGB;
    else Stream.Position:=Stream.Position+Chunk.Size-SizeOf(T3DSChunk); // skip chunk
   end;
  end;
 end;

function T3DSReader.LoadTransparency:single;
 var
  size:integer;
 begin
  Result:=0;
  Size:=Chunk.Size-SizeOf(T3DSChunk);
  while Size>0 do begin
   case LoadChunk(Size) of
    $0030 : Result:=LoadWord/100;
    else Stream.Position:=Stream.Position+Chunk.Size-SizeOf(T3DSChunk); // skip chunk
   end;
  end;
 end;

procedure T3DSReader.LoadObject;
 var
  s:string;
 begin
  s:=LoadString;
  group:=Groups.Add(s);
  Load(s);
  TranslateGroup(Matrix.Position);
 end;

procedure T3DSReader.LoadVertices;
 var
  count:word;
 begin
  vertex0:=VertexCount; // for LoadFaces
  count:=LoadWord;
  SetLength(Vertices,VertexCount+count);
  Stream.ReadBuffer(Vertices[VertexCount],count*sizeof(TVertex));
  Inc(VertexCount,Count);
 end;

procedure T3DSReader.LoadFaces;
 var
  count:word;
  items:array of T3DSFace;
  i:integer;
 begin
  face0:=FaceCount;
  count:=LoadWord;
  SetLength(items,count);
  Stream.ReadBuffer(items[0],count*Sizeof(T3DSFace));
  SetLength(Faces,FaceCount+count);
  for i:=0 to count-1 do begin
   Faces[FaceCount+i].a:=vertex0+items[i].a;
   Faces[FaceCount+i].b:=vertex0+items[i].b;
   Faces[FaceCount+i].c:=vertex0+items[i].c;
   Faces[FaceCount+i].group:=group;
   Faces[FaceCount+i].material:=-1;
  end;
  Inc(FaceCount,Count);
  Dec(Chunk.Size,Count*SizeOf(T3DSFace)+2);
  Load('Face');
 end;

procedure T3DSReader.LoadMaterials;
 var
  name:string;
  index:integer;
  count:word;
  i:integer;
 begin
  name:=LoadString;
  index:=Length(Materials)-1;
  while (index>=0)and(Materials[Index].Name<>Name) do dec(Index);
  count:=LoadWord;
  for i:=0 to count-1 do Faces[Face0+LoadWord].Material:=Index;
 end;

procedure T3DSReader.LoadMappings;
 var
  count:word;
 begin
  count:=LoadWord;
  if vertex0+count>VertexCount then raise Exception.Create('Load3DS internal error on "LoadMappings"');
  SetLength(fMappings,VertexCount);
  Stream.ReadBuffer(fMappings[vertex0],count*sizeof(T3DSMapping));
 end;

procedure T3DSReader.LoadMeshMatrix;
 begin
  Stream.ReadBuffer(matrix,sizeof(matrix));
(*
  matrix.position.x:=-matrix.position.x;
  matrix.position.y:=-matrix.position.y;
  matrix.position.z:=-matrix.position.z;
*)
 end;

procedure T3DSReader.LoadNodeHeader;
 begin
  Group:=Groups.IndexOf(LoadString);
  Stream.Position:=Stream.Position+Chunk.Size-SizeOf(T3DSChunk); // skip chunk
 {$ifdef log}write('NodeHeader[',Group,'] '); if Group>=0 then write(Groups[Group]); writeln;{$endif}
 end;

procedure T3DSReader.LoadKeyHeader(out Header:T3DSKeyHeader);
 begin
  Stream.ReadBuffer(Header,sizeof(integer)+sizeof(word));
  dec(Chunk.Size,sizeof(integer)+sizeof(word));
  with Header do begin
   if (Flags and  1)<>0 then Tension:=LoadSingle;
   if (Flags and  2)<>0 then Continuity:=LoadSingle;
   if (Flags and  4)<>0 then Bias:=LoadSingle;
   if (Flags and  8)<>0 then EaseTo:=LoadSingle;
   if (Flags and 16)<>0 then EaseFrom:=LoadSingle;
  end;
 end;

procedure T3DSReader.LoadPivot;
 begin
  Stream.ReadBuffer(Pivot,SizeOf(Pivot));
 end;

procedure T3DSReader.LoadTranslation;
 var
  Track:T3DSTrackHeader;
  Key  :T3DSKeyHeader;
  Point:TVertex;
 begin
  if Group>=0 then begin
   Stream.ReadBuffer(Track,sizeof(Track));
   dec(Chunk.size,sizeof(Track));
   LoadKeyHeader(Key);
   Stream.ReadBuffer(Point,SizeOf(Point));
   dec(Chunk.Size,sizeof(Point));
  {
   point.x:=-point.x;
   point.y:=-point.y;
   point.z:=-point.z;
  }
   TranslateGroup(point);
   writeln('frame translation [',group,'] : ',point.x:4:2,',',point.y:4:2,',',point.z:4:2);
  end;
  Stream.Position:=Stream.Position+Chunk.Size-SizeOf(T3DSChunk); // skip chunk
 end;

procedure T3DSReader.LoadRotation;
 var
  Track :T3DSTrackHeader;
  Key   :T3DSKeyHeader;
  angle :single;
  vector:TVector;
  i:integer;
  p:array of byte;
  sn:single;
  cs,one_cs:single;
  xx,yy,zz:single;
  xy,yz,zx:single;
  xs,ys,zs:single;
  tx,ty,tz:single;
 begin
  Stream.ReadBuffer(Track,sizeof(Track));
  dec(Chunk.size,sizeof(Track));
  LoadKeyHeader(Key);
  Angle:=LoadSingle;
  if Angle<>0 then begin
   writeln('rotation');
   Stream.ReadBuffer(Vector,sizeof(Vector));
   dec(Chunk.size,sizeof(Vector));
   SetLength(p,VertexCount);
   for i:=0 to FaceCount-1 do begin
    if Faces[i].group=Group then begin
     with Faces[i] do begin
      p[a]:=1;
      p[b]:=1;
      p[c]:=1;
     end;
    end;
   end;
  // perform rotation of this points
   angle:=angle*PI/180;
   sn:=sin(angle);
   cs:=cos(angle);
   one_cs:=1-cs;
   with Vector do begin
    xx:=x*x;
    yy:=y*y;
    zz:=z*z;
    xy:=x*y;
    yz:=y*z;
    zx:=z*x;
    xs:=x*sn;
    ys:=y*sn;
    zs:=z*sn;
   end;
   for i:=0 to VertexCount-1 do if p[i]=1 then with Vertices[i] do begin
    x:=x+Pivot.x;
    tx:=x*((one_cs*xx)+cs);
    ty:=x*((one_cs*xy)-zs);
    tz:=x*((one_cs*zx)+ys);

    y:=y+Pivot.y;
    tx:=tx+y*((one_cs*xy)+zs);
    ty:=ty+y*((one_cs*yy)+cs);
    tz:=tz+y*((one_cs*yz)-xs);

    z:=z-Pivot.z;
    x:=tx+z*((one_cs*zx)-ys);
    y:=ty+z*((one_cs*yz)+xs);
    z:=tz+z*((one_cs*zz)+cs);
   end;
   p:=nil;
  end;
  Stream.Position:=Stream.Position+Chunk.Size-SizeOf(T3DSChunk); // skip chunk
 end;

procedure T3DSReader.LoadScale;
 var
  Track:T3DSTrackHeader;
  Key  :T3DSKeyHeader;
  Scale:TVector;
  i:integer;
  p:array of byte;
 begin
  if Group>=0 then begin
   Stream.ReadBuffer(Track,sizeof(Track));
   dec(Chunk.size,sizeof(Track));
   LoadKeyHeader(Key);
   Stream.ReadBuffer(Scale,SizeOf(Scale));
   dec(Chunk.size,sizeof(Scale));
   if (scale.x<>1)or(scale.y<>1)or(scale.z<>1) then begin
   writeln('echelle');
    SetLength(p,VertexCount);
    for i:=0 to FaceCount-1 do begin
     if Faces[i].group=Group then begin
      with Faces[i] do begin
       p[a]:=1;
       p[b]:=1;
       p[c]:=1;
      end;
     end;
    end;
    for i:=0 to VertexCount-1 do begin
     if p[i]=1 then with Vertices[i] do begin
      x:=x*Scale.x;
      y:=y*Scale.y;
      z:=z*Scale.z;
     end;
    end;
    p:=nil;
   end;
  end;
  Stream.Position:=Stream.Position+Chunk.Size-SizeOf(T3DSChunk); // skip chunk
 end;

procedure T3DSReader.TranslateGroup(const v:TVector);
 var
  i:integer;
  p:array of integer;
 begin
  if group<0 then exit;
  if (v.x<>0)or(v.y<>0)or(v.z<>0) then begin
   SetLength(p,VertexCount);
   for i:=0 to FaceCount-1 do begin
    if Faces[i].group=Group then begin
     with Faces[i] do begin
      p[a]:=1;
      p[b]:=1;
      p[c]:=1;
     end;
    end;
   end;
   for i:=0 to VertexCount-1 do begin
    if p[i]=1 then with Vertices[i] do begin
     x:=x+v.x;
     y:=y+v.y;
     z:=z+v.z;
    end;
   end;
   p:=nil;
  end;
 end;

function T3DSReader.AddTexture(AFileName:string):integer;
 begin
  AFileName:=copy(AFileName,1,Length(AFileName)-3)+'BMP';
  Result:=TextureCount;
  inc(TextureCount);
  SetLength(Textures,TextureCount);
  Textures[Result]:=TTexture.Create;
  Textures[Result].FileName:=AFileName;
 end;

end.

